    if (probefunc == "posix_spawn" ||
        probefunc == "posix_spawnp" ||
        probefunc == "clone" ||
        probefunc == "__clone2" ||
        probefunc == "clone3" ||
        probefunc == "fork" ||
        probefunc == "vfork" ||
        probefunc == "forkexec" || 
        probefunc == "execl" || 
        probefunc == "execlp" || 
        probefunc == "execle" || 
        probefunc == "execv" || 
        probefunc == "execvp")